import json
import re
import torch
from collections import Counter
from transformers import AutoModelForCausalLM, AutoTokenizer

MODEL_NAME = "microsoft/phi-2"
DATASET_PATH = "path/to/your/dataset.jsonl"

def format_prompt(question, choices):
    prompt = question["stem"] + " "
    prompt += " ".join([f"{choice['label']}) {choice['text']}" for choice in question["choices"]])
    return prompt

def get_model_answer(model, tokenizer, prompt):
    inputs = tokenizer(prompt, return_tensors="pt", return_attention_mask=True)
    outputs = model.generate(**inputs, max_length=250)
    answer = tokenizer.batch_decode(outputs)[0]

    answer_letter = re.search(r"Answer:\s*([A-D])\)", answer)
    return answer_letter.group(1) if answer_letter else "No proper response"

def evaluate_model(model_name, dataset_path):
    torch.set_default_device("cuda")
    model = AutoModelForCausalLM.from_pretrained(model_name, torch_dtype="auto", trust_remote_code=True)
    tokenizer = AutoTokenizer.from_pretrained(model_name, trust_remote_code=True)

    correct_answers = 0
    total_questions = 0
    answer_distribution = Counter()

    with open(dataset_path, 'r') as file:
        for line in file:
            data = json.loads(line)
            question = data['question']
            answer_key = data['answerKey']
            prompt = format_prompt(question, question["choices"])
            model_answer = get_model_answer(model, tokenizer, prompt)

            total_questions += 1
            if model_answer == answer_key:
                correct_answers += 1
            answer_distribution[model_answer] += 1
            print(f"Question ID: {data['id']} | Model Answer: {model_answer} | Correct: {model_answer == answer_key}")

    accuracy_rate = (correct_answers / total_questions) * 100 if total_questions > 0 else 0
    print(f"Accuracy Rate: {accuracy_rate:.2f}%")
    print(f"Total Questions: {total_questions}")
    print(f"Correct Answers: {correct_answers}")
    print(f"Answer Distribution: {answer_distribution}")

if __name__ == "__main__":
    evaluate_model(MODEL_NAME, DATASET_PATH)