from transformers import (
    LlamaTokenizer,
    LlamaForCausalLM,
    PreTrainedModel,
    PreTrainedTokenizer,
)
from core import filter_code, run_eval, fix_indents
import os
import torch


@torch.inference_mode()
def generate_batch_completion(
    model: PreTrainedModel, tokenizer: PreTrainedTokenizer, prompt, batch_size
) -> list[str]:
    input_batch = [prompt for _ in range(batch_size)]
    # Adjusted to use "cuda" explicitly, assuming your setup is using CUDA devices.
    inputs = tokenizer(input_batch, return_tensors="pt").to("cuda")
    input_ids_cutoff = inputs.input_ids.size(dim=1)

    generated_ids = model.module.generate(  # Use `.module` to access the original model attributes if needed.
        **inputs,
        use_cache=True,
        max_new_tokens=512,
        temperature=0.2,
        top_p=0.95,
        do_sample=True,
        eos_token_id=tokenizer.eos_token_id,
        pad_token_id=tokenizer.eos_token_id,  # model has no pad token
    )

    batch_completions = tokenizer.batch_decode(
        [ids[input_ids_cutoff:] for ids in generated_ids],
        skip_special_tokens=True,
    )

    return [filter_code(fix_indents(completion)) for completion in batch_completions]



if __name__ == "__main__":
    # adjust for n = 10 etc
    num_samples_per_task = 25
    out_path = "results/llama/eval.jsonl"
    os.makedirs("results/llama", exist_ok=True)

    tokenizer = LlamaTokenizer.from_pretrained(
        "huggyllama/llama-7b",
    )

    model = model = LlamaForCausalLM.from_pretrained(
        "huggyllama/llama-7b",
        torch_dtype=torch.bfloat16,
    ).eval()
    
    model = torch.nn.DataParallel(model)
    model.to("cuda")

    run_eval(
        model,
        tokenizer,
        num_samples_per_task,
        out_path,
        generate_batch_completion,
        True,
    )
