setups = {
    "35M": { "L": 6, "H": 8, "D": 512, },
    "70M": { "L": 10, "H": 10, "D": 640, },
    "284M": { "L": 20, "H": 16, "D": 1024, },
    "512M": { "L": 24, "H": 10, "D": 1280, },
    "1B": { "L": 26, "H": 14, "D": 1792, },
    "1.5B": { "L": 28, "H": 16, "D": 2048, },
    "6.5B": { "L": 32, "H": 32, "D": 4096, },
    "13B": { "L": 40, "H": 40, "D": 5120, },
    "30B": { "L": 60, "H": 52, "D": 6656, },
    "65B": { "L": 80, "H": 64, "D": 8192, },
    "140B": { "L": 80, "H": 96, "D": 12288, },
    "310B": { "L": 96, "H": 128, "D": 16384, },
    "1T": { "L": 128, "H": 160, "D": 25600, }
}

CURRENT = setups["65B"]

L = CURRENT["L"] # number of layers
H = CURRENT["H"] # number of heads
D = CURRENT["D"] # embedding dimension

BS = 1 # batch size
SEQ = 4096 # sequence length
TOKS = 32_000 # number of tokens in the vocab

# optimizer momentums saved per parameter (Adam holds 2, SGD holds 1)
OPTIMIZER_MOMENTUMS = 2

# fp16 percent memory saved for activations; realistic values range beetwen 15%
# and 30%; for full fp32 training, set to 0
FP16_SAVED_ACTIVATIONS = 0

# -- END OF GLOBALS --

bytes_to_gigs = 1_073_741_824 # bytes in a gigabyte

# model parameters
embedding_layer = TOKS * D
multi_head_attention_layer = 4 * D * D
feed_forward_layer = 3 * D * (8 * D // 3)
norm_layer = D
out_layer = TOKS * D
model_params = embedding_layer + L * (multi_head_attention_layer + \
    feed_forward_layer + 2 * norm_layer) + norm_layer + out_layer

# activations
FFN_DIM = 8 * D // 3
activations = (BS * SEQ + # input embedding
    \
    \
    # attention layer + norm
    L * (
        3 * BS * SEQ * D + # (BS, SEQ, D)
        2 * BS * SEQ * 1 + # (BS, SEQ, 1)
        4 * BS * SEQ * D + # (BS * SEQ, D)
        4 * D * D + # (D, D)
        1 * BS * H * SEQ * D / H + # (BS * H, SEQ, D / H)
        1 * BS * H * D / H * SEQ + # (BS * H, D / H, SEQ)
        1 * BS * H * SEQ * SEQ + # (BS, H, SEQ, SEQ)
        1 * BS * H * SEQ * SEQ + # (BS * H, SEQ, SEQ)
        1 * BS * H * SEQ * D / H # (BS * H, SEQ, D / H)
    ) +
    \
    \
    # ffn layer + norm
    L * (
        3 * BS * SEQ * D + # (BS, SEQ, D)
        2 * BS * SEQ * 1 + # (BS, SEQ, 1)
        2 * BS * SEQ * D + # (BS * SEQ, D)
        2 * D * FFN_DIM + # (D, FFN_DIM)
        3 * BS * SEQ * FFN_DIM + # (BS, SEQ, FFN_DIM)
        1 * BS * SEQ * FFN_DIM + # (BS * SEQ, FFN_DIM)
        1 * FFN_DIM * D # (FFN_DIM, D)
    ) +
    \
    \
    # output layer + norm
    1 * BS * SEQ * D + # (BS * SEQ, D)
    3 * BS * SEQ * D + # (BS, SEQ, D)
    2 * BS * SEQ * 1 + # (BS, SEQ, 1)
    1 * D * TOKS # (D, TOKS)
)

# backpropagation gradients
gradients = 1 * model_params

# optimizer state
optimizer = OPTIMIZER_MOMENTUMS * model_params

# 4 bytes (fp32) used; for 2 bytes activations (fp16), adjust the percent value;
gigabytes_used = 4 * (
    model_params +
    optimizer +
    (1 - FP16_SAVED_ACTIVATIONS) * activations +
    gradients
) / bytes_to_gigs

print(f"Model params: {model_params:,} - {model_params * 4 / bytes_to_gigs:,.2f}GB")
print(f"Activations: {activations:,} - {activations * 4 / bytes_to_gigs:,.2f}GB")
print(f"Backprop gradients: {gradients:,} - {gradients * 4 / bytes_to_gigs:,.2f}GB")
print(f"GB total (no parallelism): {gigabytes_used:,.2f}GB")
print(f"GB total (+40% memory overhead for tensor and pipeline parallelism): {1.4 * gigabytes_used:,.2f}GB")
print(f"Minimum A100 (80GB) GPUs (with overhead): {1.4 * (gigabytes_used / 80):.1f}")
